/*
 * UOT-100 ISDB-T One Segment Tuner driver
 *
 * kohei sekine (eggman@ptie.org)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, version 2.
 */

//#define DEBUG
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/delay.h>
#include <asm/uaccess.h>

#include "uot100.h"

#define USB_UOT100_VENDOR_ID	0x10c4
#define USB_UOT100_PRODUCT_ID	0x2030

#define SET_REG_DELAY 10

static struct usb_device_id uot100_table [] = {
/* 00 */	{ USB_DEVICE(USB_UOT100_VENDOR_ID, USB_UOT100_PRODUCT_ID) },
		{ }		/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, uot100_table);

/* temporary use */
#define USB_UOT100_MINOR_BASE	192

struct usb_uot100 {
	struct usb_device *	udev;
	struct usb_interface *	interface;
	unsigned char *		bulk_in_buffer;
	size_t			bulk_in_size;
	__u8			bulk_in_endpoint_addr;
	__u8			bulk_out_endpoint_addr;
	struct kref		kref;
	struct urb		ctrl_urb;
	struct usb_ctrlrequest  dr;
	unsigned char		reg[20];
	int			channel;		/* curent channel */
};

#define to_uot100_dev(d)	container_of(d, struct usb_uot100, kref)

static struct usb_driver uot100_driver;

/*********************/
/* internal function */
/*********************/
static void uot100_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
{
	if(urb->status &&
	   !(urb->status == -ENOENT ||
	     urb->status == -ECONNRESET ||
	     urb->status == -ESHUTDOWN)){
		err("%s - nonzero write bulk status received: %d", __FUNCTION__, urb->status);
	}

	usb_buffer_free(urb->dev, urb->transfer_buffer_length, urb->transfer_buffer, urb->transfer_dma);
}

static void uot100_ctrl_callback(struct urb *urb)
{
	if(urb->status &&
	   !(urb->status == -ENOENT ||
	     urb->status == -ECONNRESET ||
	     urb->status == -ESHUTDOWN)){
		err("%s - nonzero write ctrl status received: %d", __FUNCTION__, urb->status);
	}

	return;
}

int uot100_set_config(struct usb_uot100 *dev, u8 reg1, u8 reg2, u8 reg3, u8 reg4, u8 reg5 , u8 reg6, u8 reg7)
{
	int retval = 0;
	
	dbg("uot100_set_config() reg1=0x%02x reg2=0x%02x reg3=0x%02x reg4=0x%02x reg5=0x%02x reg6=0x%02x reg7=0x%02x ",
	    reg1, reg2, reg3, reg4, reg5, reg6, reg7);

        /* send vender device packet */
	dev->dr.bRequestType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	dev->dr.bRequest     = reg1;
	dev->dr.wValue       = cpu_to_le16(reg2 + (reg3<<8));
	dev->dr.wIndex       = cpu_to_le16(reg4 + (reg5<<8));
	dev->dr.wLength      = cpu_to_le16(reg6 + (reg7<<8));
	usb_fill_control_urb (&dev->ctrl_urb, dev->udev,
				usb_sndctrlpipe(dev->udev, 0),
				(unsigned char *)&dev->dr,
				&reg1, 0,
				(usb_complete_t )uot100_ctrl_callback,
				dev);

	retval = usb_submit_urb(&dev->ctrl_urb, GFP_KERNEL);
	if(retval){
		err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		return retval;
	}

	mdelay(SET_REG_DELAY);
	return retval;
}

#define REG_BUF_SIZE 5
int uot100_write_reg(struct usb_uot100 *dev, u8 reg1, u8 reg2, u8 reg3, u8 reg4, u8 reg5 )
{
	u8 *buf=NULL;
	struct urb *urb=NULL;
	int len = REG_BUF_SIZE, retval;

	//dbg("uot100_write_reg() reg1=0x%02x reg2=0x%02x reg3=0x%02x reg4=0x%02x reg5=0x%02x",
	//    reg1, reg2, reg3, reg4, reg5);

	urb = usb_alloc_urb(0, GFP_KERNEL);
	if(!urb){
		retval = -ENOMEM;
		goto error;
	}
	buf = usb_buffer_alloc(dev->udev, REG_BUF_SIZE, GFP_KERNEL, &urb->transfer_dma);
	if(!buf){
		retval = -ENOMEM;
		goto error;
	}
	buf[0] = reg1;
	buf[1] = reg2;
	buf[2] = reg3;
	buf[3] = reg4;
	buf[4] = reg5;

	usb_fill_bulk_urb(urb, dev->udev,  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpoint_addr),
	                  buf, REG_BUF_SIZE, (usb_complete_t)uot100_write_bulk_callback, dev);
	urb->transfer_flags = 0x2;

	retval = usb_submit_urb(urb, GFP_KERNEL);
	if(retval){
		err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	usb_free_urb(urb);

	return 0;
error:
	usb_buffer_free(dev->udev, len, buf, urb->transfer_dma);
	usb_free_urb(urb);
	kfree(buf);
	return retval;
}

#define READ_REG_SIZE 20
static void uot100_non_block_read_reg_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_uot100 *dev = urb->context;

	if(urb->status &&
	   !(urb->status == -ENOENT ||
	     urb->status == -ECONNRESET ||
	     urb->status == -ESHUTDOWN)){
		err("%s - nonzero read bulk status received: %d", __FUNCTION__, urb->status);
		memset(&dev->reg, 0x00, READ_REG_SIZE);
	}
	else{
#ifdef DEBUG
		unsigned char *buf = urb->transfer_buffer;
		info("nbread  %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
		      *buf,*(buf+1),*(buf+2),*(buf+3),*(buf+4),
		      *(buf+5),*(buf+6),*(buf+7),*(buf+8),*(buf+9),
		      *(buf+10),*(buf+11),*(buf+12),*(buf+13),*(buf+14),
		      *(buf+15),*(buf+16),*(buf+17),*(buf+18),*(buf+19)
		      );
#endif
		memcpy(&dev->reg, urb->transfer_buffer, READ_REG_SIZE);
	}

	//atode
	//usb_buffer_free(urb->dev, urb->transfer_buffer_length, urb->transfer_buffer, urb->transfer_dma);
}

int uot100_non_block_read_reg(struct usb_uot100 *dev)
{
	int retval = 0;
	struct urb *urb=NULL;

	dbg("uot100_non_block_read_reg %d", READ_REG_SIZE);
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if(!urb){
		retval = -ENOMEM;
		goto exit;
	}

	usb_fill_bulk_urb(urb, dev->udev,  usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpoint_addr),
	                  dev->bulk_in_buffer, READ_REG_SIZE, (usb_complete_t)uot100_non_block_read_reg_callback, dev);
	urb->transfer_flags =  0x3;

	retval = usb_submit_urb(urb, GFP_KERNEL);
	if(retval){
		err("%s - failed submitting non block read urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	retval = 0;
exit:
	return retval;
error:
	usb_free_urb(urb);
	return retval;
}

int uot100_non_block_read_ts(struct usb_uot100 *dev)
{
	int retval = 0;
	struct urb *urb=NULL;

	dbg("uot100_non_block_read_reg %d", READ_REG_SIZE);
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if(!urb){
		retval = -ENOMEM;
		goto exit;
	}

	usb_fill_bulk_urb(urb, dev->udev,  usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpoint_addr),
	                  dev->bulk_in_buffer, 0xc5, (usb_complete_t)uot100_non_block_read_reg_callback, dev);
	urb->transfer_flags =  0x3;

	retval = usb_submit_urb(urb, GFP_KERNEL);
	if(retval){
		err("%s - failed submitting non block read urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	retval = 0;
exit:
	return retval;
error:
	usb_free_urb(urb);
	return retval;
}


int uot100_set_reg(struct usb_uot100 *dev, u16 addr, u8 data )
{
	int retval =0;
	dbg("uot100_set_reg() addr=0x%04x data=0x%02x", addr, data);
	retval = uot100_write_reg(dev,
				0,
				addr >> 8,
				addr & 0xff,
				data,
				0);
	mdelay(SET_REG_DELAY);
	return retval;
}

int uot100_get_reg(struct usb_uot100 *dev, u16 addr, u8 *data )
{
	int len = REG_BUF_SIZE;
	int retval;

	dbg("uot100_get_reg() addr=0x%04x",addr);

	/* set register addres */ 
	retval = uot100_write_reg(dev,
				1,
				addr >> 8,
				addr & 0xff,
				0xff,
				0xff);

	dbg("uot100_set_reg() retval=%d len=%d", retval, len);
	if(retval != 0 || len <  REG_BUF_SIZE)
		return -EFAULT;
	
	/* delay */
	mdelay(SET_REG_DELAY);

	/* read last read */
	if(dev->reg[0] == 0x48 && dev->reg[1] == 0x75 && dev->reg[2] == 0x09 &&
	   dev->reg[3] == 0x15 && dev->reg[4] == 0x58 && dev->reg[5] == 0x01 && dev->reg[6] == 0x01){
		*data = dev->reg[7];
		retval = 0;
	}else
		retval = -EFAULT;
	
	/* get register for next */
	uot100_non_block_read_reg(dev);
	
	mdelay(SET_REG_DELAY);
	return retval;
}

int uot100_reset(struct usb_uot100 *dev)
{
	unsigned char data = 0;
	int i, count = 5, flag = 0;

	const u8 boot1[] = { 	0x00, 0x70, 0x89, 0xc3, 0x12, 0x83, 0x00, 0x20,
				0x00, 0x27, 0x14, 0x04, 0xb0, 0x00, 0x93, 0x30,
				0x09, 0x1d, 0x8c, 0x9c, 0xa4, 0x09, 0xff, 0xb2,
				0x94, 0xce, 0x55, 0x3a, 0x7f, 0x81, 0x7f, 0x81,
				0x01, 0xc6, 0x72, 0x46, 0x7e, 0x13, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xe0, 0x7f, 0x04, 0xb0, 0x7e,
				0x7f, 0x02, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f,
				0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f,
				0x1f, 0x1f, 0x21, 0x02, };
	const u16 boot2[26][2] = {{0x17f, 0xb0}, {0x174, 0x70}, {0x101, 0x01}, {0x005, 0x01}, {0x006, 0x70},
			          {0x04d, 0x0e}, {0x070, 0x01}, {0x040, 0x0f}, {0x043, 0xf8}, {0x047, 0xc0},
			          {0x0a1, 0x40}, {0x0a5, 0x4c}, {0x0b9, 0xd9}, {0x0ac, 0x01}, {0x0ad, 0x80},
			          {0x0ae, 0x43}, {0x0af, 0x01}, {0x0b0, 0x22}, {0x0bd, 0xe9}, {0x0c0, 0x1c},
			          {0x0c4, 0x20}, {0x0c5, 0x04}, {0x0ca, 0x33}, {0x008, 0x0c}, {0x009, 0x12},
			          {0x001, 0x02}, };
	const u16 boot3[6][2] =  {{0x175, 0x89}, {0x178, 0x83}, {0x179, 0x00}, {0x17a, 0x20}, {0x17b, 0x00},
			          {0x17d, 0x14}, };

	dbg("uot100_reset");
	while(flag == 0 && count > 0){
		uot100_set_config(dev, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00);	
		uot100_non_block_read_reg(dev);
		for(i = 0x173; i <= 0x1b6; i++) {
			if(uot100_set_reg(dev, i, boot1[i-0x173]) != 0) return -EFAULT;
		}
        
		if(uot100_set_reg(dev, 0x1a2, 0x00)  != 0) return -EFAULT;
		if(uot100_set_reg(dev, 0x1a3, 0x48)  != 0) return -EFAULT;
		if(uot100_get_reg(dev, 0x17f, &data) != 0) return -EFAULT;
		uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);	
		if(uot100_set_reg(dev, 0x17f, 0x60)  != 0) return -EFAULT;
		if(uot100_get_reg(dev, 0x168, &data) != 0) return -EFAULT;
		uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);	
        
		for(i = 0; i < 26; i++) {
			if(uot100_set_reg(dev, boot2[i][0], boot2[i][1]) != 0) return -EFAULT;
		}
		mdelay(100);
		for(i = 0; i < 5; i++) {
			if(uot100_set_reg(dev, boot3[i][0], boot3[i][1]) != 0) return -EFAULT;
		}
		if(dev->reg[0] == 0x48)
			flag = 1;
		count--;
	}
	return 0;
}

/******************/
/* ioctl function */
/******************/
int uot100_set_channel(struct usb_uot100 *dev, int channel)
{
	unsigned char p1, p2, p3, p4, p5, p2bef, data;
	int freq, count, flag, tune, retval;

		/* 0    1    2    3    4    5    6    7    8    9*/
	const unsigned char p1a[] = {
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x88,0x89,0x89,0x8A,0x8A,0x8A,0x8B,
		0x8B,0x8C,0x8C,0x8D,0x8D,0x8D,0x8E,0x8E,0x8F,0x8F,
		0x86,0x86,0x87,0x87,0x88,0x88,0x89,0x89,0x8A,0x8A,
		0x8A,0x8B,0x8B,0x8C,0x8C,0x8C,0x8D,0x8D,0x8D,0x8E,
		0x8E,0x8F,0x8F,0x86,0x86,0x86,0x87,0x87,0x87,0x88,
		0x88,0x89,0x8A };
	const unsigned char p2a[] = {
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x30,0x30,0x30,0x30,0x30,0x30,0x30,
		0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30,
		0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,
		0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,
		0x70,0x70,0x70,0xB0,0xB0,0xB0,0xB0,0xB0,0xB0,0xB0,
		0xB0,0xB0,0xB0 };

	dbg("uot100_set_channel channel=%02d", channel);
	if(channel <= 13 || 62 <= channel)
		return -EFAULT;
	
	dev->channel = 0;
	retval = uot100_reset(dev);
	if(retval != 0)
		return -EFAULT;

	dbg("uot100_set_channel tuning");

	/* set parametor */
	freq = 473 + (channel-13)*6;
	p1 = p1a[channel];
	p2 = p2a[channel];
	p3 = freq >> 4;
	p4 = freq & 0x0f;
	if( channel <= 20)
		p5 = 0x8c;
	else
		p5 = 0x94;
	p2bef = 0;
	count = 10;
	flag = 0;



	uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x0f, 0x88);	
	/* set channel */
	while(flag == 0 && count > 0){
		mdelay(50);
		dbg("p2=0x%02x p2bef=0x%02x", p2, p2bef);
		if(p2 != p2bef){
			uot100_set_reg(dev, 0x178, 0x83);
			uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x0f, 0x88);	
			uot100_set_reg(dev, 0x175, p1);
			uot100_set_reg(dev, 0x182, p2);
			uot100_set_reg(dev, 0x184, p3);
			uot100_set_reg(dev, 0x185, p5);
			uot100_set_reg(dev, 0x186, 0x9c);
			uot100_set_reg(dev, 0x187, 0xa4);
			uot100_set_reg(dev, 0x188, p4);
			uot100_set_reg(dev, 0x183, 0x0b);
			uot100_set_reg(dev, 0x183, 0x09);
			p2bef = p2;	
		}else{
			uot100_set_reg(dev, 0x175, p1);
			uot100_set_reg(dev, 0x182, p2);
			p2bef = p2;	
		}	
		mdelay(50);
		uot100_get_reg(dev, 0x170, &data);
		mdelay(20);
		uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);	
		tune = (data >> 4) & 0x0f;

		dbg("tune=0x%02x", tune);
		/* check tuning level */
		if(tune<3){
			p1 = p1 -1;
			if(p1 < 0x86){
				p2 = ((((p2 >>5)-1) <<5) | 0x10);
			}
		}else{
			if(tune > 0x08){
				p1 = p1 +1;
				if(p1 > 0x8f){
					p2 = ((((p2 >>5)+1) <<5) | 0x10);
				}
			}else{
				flag = 1;
			}
		}
		count--;
	}
	mdelay(10);
	uot100_get_reg(dev, 0x169, &data);
	uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);	
	uot100_set_reg(dev, 0x175, p1);
	uot100_set_reg(dev, 0x001, 0x02);
	uot100_get_reg(dev, 0x032, &data);
	uot100_set_config(dev, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);	
	if((data & 0x02) != 0){
		dbg("tuning ok!");
		uot100_write_reg(dev, 0x02, 0x00, 0x00, 0x00, 0x00);
	}else{
		//atode	
		uot100_write_reg(dev, 0x02, 0x00, 0x00, 0x00, 0x00);
	}
	uot100_non_block_read_ts(dev);
	return 0;
}

/***************************/
/* File Operation function */
/***************************/
static void uot100_delete(struct kref *kref)
{
	struct usb_uot100 *dev = to_uot100_dev(kref);
	
	usb_put_dev(dev->udev);
	kfree(dev->bulk_in_buffer);
	kfree(dev);
}

static int uot100_open(struct inode *inode, struct file *file)
{
	struct usb_uot100 *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;

	subminor = iminor(inode);

	interface = usb_find_interface(&uot100_driver, subminor);
	if(!interface){
		err("%s - error, can't find device for minor %d", __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if(!dev){
		retval = -ENODEV;
		goto exit;
	}

	kref_get(&dev->kref);

	file->private_data = dev;
exit:
	dbg("open %d\n", retval);
	return retval;
}

static int uot100_release(struct inode *inode, struct file *file)
{
	struct usb_uot100 *dev;

	dev = (struct usb_uot100 *) file->private_data;
	if(dev == NULL)
		return -ENODEV;
	
	kref_put(&dev->kref, uot100_delete);
	return 0;
}

static ssize_t uot100_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	struct usb_uot100 *dev;
	int retval = 0;
	
	dev = (struct usb_uot100 *) file->private_data;

	dbg("read count %d", count);
	retval = usb_bulk_msg(dev->udev,
			      usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpoint_addr),
			      dev->bulk_in_buffer,
			      min(dev->bulk_in_size, count),
			      &count, HZ*10);
	
	if(!retval){
		if(copy_to_user(buffer, dev->bulk_in_buffer, count))
			retval = -EFAULT;
		else{
#ifdef DEBUG
			unsigned char *buf = dev->bulk_in_buffer;
			info("read len:%d %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
			      count,
			      *buf,*(buf+1),*(buf+2),*(buf+3),*(buf+4),
			      *(buf+5),*(buf+6),*(buf+7),*(buf+8),*(buf+9),
			      *(buf+10),*(buf+11),*(buf+12),*(buf+13),*(buf+14),
			      *(buf+15),*(buf+16),*(buf+17),*(buf+18),*(buf+19)
			      );
#endif
			retval = count;
		}
	}else{
		dbg("can't read\n");
	}

	return retval;
}

static ssize_t uot100_write(struct file *file, const char __user *user_buffer, size_t count, loff_t *ppos)
{
	struct usb_uot100 *dev;
	int retval = 0;
	struct urb *urb=NULL;
	unsigned char *buf=NULL;


	dev = (struct usb_uot100 *) file->private_data;

	if(count == 0)
		goto exit;

	urb = usb_alloc_urb(0, GFP_KERNEL);
	if(!urb){
		retval = -ENOMEM;
		goto error;
	}

	buf = usb_buffer_alloc(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
	if(!buf){
		retval = -ENOMEM;
		goto error;
	}
	if(copy_from_user(buf, user_buffer, count)){
		retval = -EFAULT;
		goto error;
	}
	dbg("write %x %x %x %x %x", *buf,*(buf+1),*(buf+2),*(buf+3),*(buf+4));

	usb_fill_bulk_urb(urb, dev->udev,  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpoint_addr),
	                  buf, (int) count, (usb_complete_t)uot100_write_bulk_callback, dev);
	//urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	urb->transfer_flags = 0x2;

	retval = usb_submit_urb(urb, GFP_KERNEL);
	if(retval){
		err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	usb_free_urb(urb);

exit:
	return count;

error:
	usb_buffer_free(dev->udev, count, buf, urb->transfer_dma);
	usb_free_urb(urb);
	kfree(buf);
	return retval;
}

int uot100_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct usb_uot100 *dev;
	int retval = 0;
	int err = 0;
	int channel;

	/* sanity check */
	if (_IOC_TYPE(cmd) != UOT100_IOC_MAGIC) return -ENOTTY;

	/* */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	if (err) return -EFAULT;
	if (_IOC_DIR(cmd) & _IOC_WRITE)
		err =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	if (err) return -EFAULT;

	dev = (struct usb_uot100 *) file->private_data;

	switch(cmd){
		case UOT100_START_STREAM:
			retval = uot100_set_channel(dev, 24);
			break;
		case UOT100_SET_CHANNEL:
			channel = (int) arg;
			retval = uot100_set_channel(dev, channel);
			break;
		case UOT100_GET_CHANNEL:
			retval = __put_user(dev->channel, (int __user *) arg);
			break;
		case UOT100_NB_READ:
		default:
			break;
	}
	return retval;
}

static struct file_operations uot100_fops = {
	.owner =	THIS_MODULE,
	.open =		uot100_open,
	.release =	uot100_release,
	.read =		uot100_read,
	.write =	uot100_write,
	.ioctl =	uot100_ioctl,
};

static struct usb_class_driver uot100_class = {
	.name = "usb/uot%d",
	.fops = &uot100_fops,
	.minor_base = USB_UOT100_MINOR_BASE,
};

static int uot100_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_uot100 *dev;
	struct usb_host_interface *if_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int retval = -ENOMEM;

	/* allocate memory */
	dev = kmalloc(sizeof(struct usb_uot100), GFP_KERNEL);
	if(dev == NULL){
		err("Out of memory");
		goto error;
	}
	memset(dev, 0x00, sizeof(*dev));
	kref_init(&dev->kref);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	/* setup endpoint information */
	if_desc = interface->cur_altsetting;
	for(i = 0; i<if_desc->desc.bNumEndpoints; ++i){
		endpoint = &if_desc->endpoint[i].desc;

		if(!dev->bulk_in_endpoint_addr &&
		   (endpoint->bEndpointAddress & USB_DIR_IN) &&
		   ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)){
			
			/* found a bulk in endpoint */
			buffer_size = endpoint->wMaxPacketSize;
			dev->bulk_in_size = buffer_size;
			dev->bulk_in_endpoint_addr = endpoint->bEndpointAddress;
			dev->bulk_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
			if(!dev->bulk_in_buffer){
				err("Could not allocate bulk_in_buffer");
				goto error;
			}
		}

		if(!dev->bulk_out_endpoint_addr &&
		   !(endpoint->bEndpointAddress & USB_DIR_IN) &&
		   ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)){
			
			/* found a bulk out endpoint */
			dev->bulk_out_endpoint_addr = endpoint->bEndpointAddress;
		}
	}
	if(!(dev->bulk_in_endpoint_addr && dev->bulk_out_endpoint_addr)){
		err("Could not find both bulk-in and bulk-out endpoints");
		goto error;
	}

	/* save data point in this interface device */
	usb_set_intfdata(interface, dev);

	retval = usb_register_dev(interface, &uot100_class);
	if(retval){
		err("Could not find both bulk-in and bulk-out endpoints");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	/* init parameter */
	usb_init_urb(&dev->ctrl_urb);
	dev->channel = 0;

	info("%s(%s) now attached to /dev/uot%d",  dev->udev->product, dev->udev->serial , interface->minor);

	return 0;
error:
	if(dev)
		kref_put(&dev->kref, uot100_delete);
	return retval;
}

static void uot100_disconnect(struct usb_interface *interface)
{
	struct usb_uot100 *dev;
	int minor = interface->minor;

	lock_kernel();

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	usb_deregister_dev(interface, &uot100_class);

	unlock_kernel();

	kref_put(&dev->kref, uot100_delete);

	info("UOT-100 #%d now discconneted", minor);
}


/* module stuff */

static struct usb_driver uot100_driver = {
	.name 		= "UOT-100 USB ISDB-T 1seg Tuner",
	.id_table 	= uot100_table,
	.probe		= uot100_probe,
	.disconnect     = uot100_disconnect,
};

static int __init uot100_module_init(void)
{
	int result;

	result = usb_register(&uot100_driver);
	if(result){
		err("usb_register failed. Error number %d",result);
	}

	return result;
}

static void __exit uot100_module_exit(void)
{
	/* deregister this driver from the USB subsystem */
	usb_deregister(&uot100_driver);
}

module_init (uot100_module_init);
module_exit (uot100_module_exit);

MODULE_AUTHOR("kohei sekine (eggman@ptie.org)");
MODULE_DESCRIPTION("uot100");
MODULE_VERSION("1.0");
MODULE_LICENSE("GPL");
